/////////////////////////////////////////////////////////////////////////////
// Name:        Scratchlib.cpp
// Purpose:
// Author:      Stefan Langhammer
// Modified by:
// Created:     4.12.2005
// Modified:    15.08.2006
// Copyright:   (c) Stefan Langhammer
// Licence:     LGPL
/////////////////////////////////////////////////////////////////////////////
// -----------------------------------------------------------------------------
//	Test.cpp - simple test program
// -----------------------------------------------------------------------------

#include "../include/DScratch.h"
#include "../include/ConsoleHelper.h"

#include "../OSCOutput.h"

#include <cstdio>
#include <cmath>
#include <stdlib.h>


// -----------------------------------------------------------------------------
//	main - program entry point
// -----------------------------------------------------------------------------

int main()
{
	static char *header_str = "\n=========== Scratchlib Example 0.4 + OSC ===========\n\n";

	printf("%s", header_str);

	DScratch *scratch = new DScratch();
	int i, iSelectedDevice, iActiveChannels;
	
	/////////////////////////////////////////////
	// check audio devices     
	/////////////////////////////////////////////
	
	int iNumDevices = scratch->GetNumDevices();
	if (iNumDevices == 0)	// no recording devices available ?
	{
		printf("Error: No recording device found!\n");
		printf("Press <Enter> to quit the application.\n");
		getchar();
		return 0;
	}
	printf("     Devices Name                              DriverType\n");
	printf("--------------------------------------------------------------\n");

	// Get device names
	string strDeviceName;
	for (i = 0; i < iNumDevices; i++)
	{
		scratch->GetDeviceName(i,strDeviceName);
		printf("%i. %s\n\n",i,strDeviceName.c_str());
	}

	//////////////////////////////////////////////////////////
	// Select device		( 1 / 6 )
	//////////////////////////////////////////////////////////
	printf("\n--------------------------------------------------------------\n");
	printf("\nPlease select a device:");
	while (true)
	{
		char buffer[5];
		int i, ch;

		for( i = 0; (i < 4) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
			buffer[i] = (char)ch;

		iSelectedDevice = atoi(buffer);

		if (iSelectedDevice >= 0 || iSelectedDevice < iNumDevices)
			break;

		printf("The selected device is invalid!\n Please try again.\n");
	}

	//////////////////////////////////////////////////////////
	// Select channels  ( 2 / 6)
	//////////////////////////////////////////////////////////
	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("\n--------------------------------------------------------------\n");
	printf("Supported input channels\n");

	// Count channels
	int iNumChannels = scratch->GetNumChannels(iSelectedDevice);
	int j = 0;
	for (i = 0; i < iNumChannels; i+=2)
	{
		printf("\t%i. Channel %i/%i\n\n", j, i+1, i+2);
		j++;
	}
	printf("\nPlease select the input channels:");
	while (true)
	{
		char buffer[5];
		int i, ch;

		for( i = 0; (i < 4) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
			buffer[i] = (char)ch;

		iActiveChannels = 2 + atoi(buffer)*2;

		if (iActiveChannels >= 0 || iActiveChannels < iNumChannels)
			break;

		printf("The selected channels are invalid!\n Please try again.\n");
	}
	//////////////////////////////////////////////////////////
	// Select vinyl   ( 3 / 6)
	//////////////////////////////////////////////////////////

	int selectedVinyl = 0;
	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("> using input Channels %i/%i\n",iActiveChannels-1,iActiveChannels);
	printf("\n--------------------------------------------------------------\n");
	printf("Supported vinyls\n\t0. Final Scratch\n\t1. Mixvibes\n\nPlease select the vinyl type:");

	while (true)
	{
		char buffer[5];
		int i, ch;

		for( i = 0; (i < 4) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
			buffer[i] = (char)ch;

		selectedVinyl = atoi(buffer);

		if (selectedVinyl == 0 || selectedVinyl == 1)
			break;

		printf("The selected vinyl type is invalid!\n Please try again.\n");
	}

	scratch->SetTimecodeVinyl(TimecodeVinyl(selectedVinyl));

	//////////////////////////////////////////////////////////
	// Enable RIAA correction ( 4 / 6)
	//////////////////////////////////////////////////////////

	bool	bEnableRIAA = false;
	int		selection	= 0;

	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("> using input Channels %i/%i\n",iActiveChannels-1,iActiveChannels);
	if		(TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_FINALSCRATCH)
		printf("> using vinyl type: FinalScratch\n");
	else if (TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_MIXVIBES)
		printf("> using vinyl type: Mixvibes\n");
	else
		printf("> using vinyl type: %i\n",selectedVinyl);
	printf("\n--------------------------------------------------------------\n");
	printf("0. Enable RIAA correction\n");
	printf("1. continue\n");

	while (true)
	{
		char buffer[5];
		int i, ch;

		for( i = 0; (i < 4) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
			buffer[i] = (char)ch;

		selection = (int)atof(buffer);
		if (selection == 0)
			bEnableRIAA = true;

		break;
	}

	if (bEnableRIAA)
		scratch->EnableRIAACorrection(true);

	//////////////////////////////////////////////////////////
	// Enable software amplifier ( 5 / 6)
	//////////////////////////////////////////////////////////

	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("> using input Channels %i/%i\n",iActiveChannels-1,iActiveChannels);
	if		(TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_FINALSCRATCH)
		printf("> using vinyl type: FinalScratch\n");
	else if (TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_MIXVIBES)
		printf("> using vinyl type: Mixvibes\n");
	else
		printf("> using vinyl type: %i\n",selectedVinyl);
	if (bEnableRIAA) printf("> using RIAA correction\n");
	printf("\n--------------------------------------------------------------\n");
	printf("0. Enable software preamp\n");
	printf("1. continue\n");

	bool bEnableSoftwarePreamp = false;
	while (true)
	{
		char buffer[5];
		int i, ch;

		for( i = 0; (i < 4) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
			buffer[i] = (char)ch;

		selection = (int)atof(buffer);
		if (selection == 0)
			bEnableSoftwarePreamp = true;

		break;
	}
	if (bEnableSoftwarePreamp)
	{
		//and calibrate software amplifier
		if (!scratch->SetActiveDevice(iSelectedDevice, iActiveChannels))
		{
			printf("Couldn't open the audio device. Press any key to quit.");
			getchar();
			return 0;
		}
		printf("\nStart playback and press <enter> to auto-adjust the volume.\n");
		getchar();
		float amplifyMultiplier = 1.0f;
		int   currentVolume     = 0;
		int	  goodVolume		= 16360; // 32767

		gotoxy(0,13);
		printf("calculate needed amplification...                                     \n");

		while(true)
		{
			scratch->SetAmplify(amplifyMultiplier);
			
			#ifdef WIN32
				Sleep(10); // ms
			#else
				usleep(10000);	
			#endif

			currentVolume = scratch->GetVolumePeak();
			if (currentVolume > goodVolume)
				break;
			else
				amplifyMultiplier += 0.5;
		}
	}

	

	//////////////////////////////////////////////////////////
	// Calibrate signal detection		( 6 / 6)
	//////////////////////////////////////////////////////////
	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("> using input Channels %i/%i\n",iActiveChannels-1,iActiveChannels);
	if		(TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_FINALSCRATCH)
		printf("> using vinyl type: FinalScratch\n");
	else if (TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_MIXVIBES)
		printf("> using vinyl type: Mixvibes\n");
	else
		printf("> using vinyl type: %i\n",selectedVinyl);
	if (bEnableRIAA)			printf("> using RIAA correction\n");
	if (bEnableSoftwarePreamp)	printf("> using software preamp\n");
	printf("\n--------------------------------------------------------------\n");

	float calibrationValue = 0.0;
	bool  bRunCalibration = false;

	if (!LoadSettings(calibrationValue))
	{
		if		(TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_FINALSCRATCH)
			calibrationValue = 1.2;
		else if (TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_MIXVIBES)
			calibrationValue = 1.07;
	}

	selection = 1;

	printf("0. run calibration\n");
	printf("1. use last calibration or default values %.2f\n",calibrationValue);
	

	while (true)
	{
		char buffer[5];
		int i, ch;

		for( i = 0; (i < 4) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
			buffer[i] = (char)ch;

		selection = (int)atof(buffer);
		if (selection == 0)
			bRunCalibration = true;

		break;
	}

	// Run calibration
	if (bRunCalibration)
	{
		if (!bEnableSoftwarePreamp)// if the audio device has not been opened before by the preamp
		{
			if (!scratch->SetActiveDevice(iSelectedDevice, iActiveChannels))
			{
				printf("Couldn't open the audio device. Press any key to quit.");
				getchar();
				return 0;
			}
		}
		printf("\nStart playback and press <enter> to run calibration.\n");
		getchar();

		int   timeleft = 60000;				// 60 sec in ms
		float bestCalibrationValue = 0.1;	// the value with the most Timecodes/second
		float testCalibrationValue = 0.1;   // the current testing value
		int   iTPS = 0;						// the detected Timecodes/second
		int   maxTimecodesPerSeconds = 0;	// the detected Timecodes/second with the bestCalibrationValue
		int	  iChecks = 0;					

		// Set calibration range begin
		scratch->SetCalibration(testCalibrationValue);

		while(true)
		{	
			#ifdef WIN32
				Sleep(50); // ms
			#else
				usleep(50000);// microseconds
			#endif

			// get Timecodes/second
				iTPS = scratch->GetTimecodesPerSecond();
			if (iTPS > maxTimecodesPerSeconds)
			{
				maxTimecodesPerSeconds	= iTPS;
				bestCalibrationValue	= testCalibrationValue;
			}

			gotoxy(0,14);
			printf("time left:  %4.2fs  current value: %5.2f  timecodes/s:      %7i\n", float(timeleft/1000.0f),  testCalibrationValue, iTPS);
			printf("                    best value:    %5.2f  best timecodes/s: %7i\n", 		bestCalibrationValue, maxTimecodesPerSeconds);

			// We should check each value at least 1.5 seconds to get a precise timecodes/s check
			// with an interval of 50ms and 30checks we get the 1.5s
			if (iChecks == 30)
			{
				iChecks = 0;
				testCalibrationValue += 0.1;
				// Set the value to check and reset the timecodes/s
				scratch->SetCalibration(testCalibrationValue);
			}
			else
				iChecks++;

			if (timeleft == 0) break;
				timeleft -= 50;
		}
		
		StoreSettings(bestCalibrationValue);
		calibrationValue = bestCalibrationValue;
		printf("\n Calibration done. Press any key to start the detection.\n");
		getchar();
	}

	scratch->SetCalibration(calibrationValue);

    //////////////////////////////////////////////////////////
	// Setup OSC
	//////////////////////////////////////////////////////////
	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("> using input Channels %i/%i\n",iActiveChannels-1,iActiveChannels);
	if		(TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_FINALSCRATCH)
		printf("> using vinyl type: FinalScratch\n");
	else if (TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_MIXVIBES)
		printf("> using vinyl type: Mixvibes\n");
	else
		printf("> using vinyl type: %i\n",selectedVinyl);
	if (bEnableRIAA)			printf("> using RIAA correction\n");
	if (bEnableSoftwarePreamp)	printf("> using software preamp\n");
	printf("\n--------------------------------------------------------------\n");
	printf("\n\n\nOSC Send setup (UDP):\n\n");


	int  osc_port = OSC_STD_PORT;
	char osc_address[32];
	bool use_default = false;

		
	// Port:
	printf("\nPort ('enter' for %i): ", OSC_STD_PORT);
	while (true)
	{
		char buffer[6];
		int i, ch;

		for( i = 0; (i < 6) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
		{
			buffer[i] = (char)ch;
		}

		if(i == 0)
		{
			osc_port = OSC_STD_PORT;
		}
		else
		{
		  buffer[i] = '\0';
		  osc_port = (int)atoi(buffer);
		}

		break;
	}
// - - - - - -
	// IP Address:
	printf("\nIP Address ('enter' for %s): ", OSC_STD_ADDRESS);
	
	while (true)
	{
		int i, ch;

		for( i = 0; (i < 12) &&  ((ch = getchar()) != EOF) 
								&& (ch != '\n'); i++ )
		{
		   osc_address[i] = (char)ch;
		}
		if(i)
		{
		  osc_address[i] = '\0';
		}
		else
		{
		  strncpy(osc_address, OSC_STD_ADDRESS, 32);
		}

		break;
	}

	// Connect!
	MyOSC osc_output(osc_address, osc_port);





	//////////////////////////////////////////////////////////
	// Start the detection !
	//////////////////////////////////////////////////////////

	SetCloseHandler();
	// if the audio device has not been opened before...
	if (!bRunCalibration && !bEnableSoftwarePreamp) 
	{
		if(!scratch->SetActiveDevice(iSelectedDevice, iActiveChannels)) // try to open it
		{
			printf("Initialization failed!\n Press any key to quit.");
			getchar();
			return 0;
		}
	}

	clearScreen();
	printf("%s", header_str);

	scratch->GetDeviceName(iSelectedDevice,strDeviceName);
	printf("%s\n",strDeviceName.c_str());
	printf("> using input Channels %i/%i\n",iActiveChannels-1,iActiveChannels);
	if		(TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_FINALSCRATCH)
		printf("> using vinyl type: FinalScratch\n");
	else if (TimecodeVinyl(selectedVinyl) == TIMECODE_VINYL_MIXVIBES)
		printf("> using vinyl type: Mixvibes\n");
	else
		printf("> using vinyl type: %i\n",selectedVinyl);
	if (bEnableRIAA)			printf("> using RIAA correction\n");
	if (bEnableSoftwarePreamp)	printf("> using software preamp\n");
	printf("> with a calibration value of: %.2f\n",calibrationValue);
	printf("> Streaming OSC data to %s:%i\n\n", osc_address, osc_port);
	printf("\n--------------------------------------------------------------\n");

	//printf("  Position  |    Speed    | Timecodes/s | Volume             |\n");
    
    printf("\nOSC Streaming running...\n");

	int	iTPS			= 0;	// Timecode per second
	int VolumePeak		= 0;
	//string strVolume	= "";
	double	dPosition	= 0.0;	// position in seconds
	double	dOldPos		= 0.0;	// ...
	double	dSpeed		= 0.0;
	double	dVolumePeak	= 0.0;
	double	diff;


	while(true)
	{
		// Dump position and speed every 1/20 second
		dPosition = scratch->GetPosition();
		diff = abs((double)dPosition - (double)dOldPos);

		iTPS = scratch->GetTimecodesPerSecond();
		dVolumePeak = scratch->GetVolumePeak() / 32767.0;
		dSpeed = scratch->GetSpeed() * 100.0;

		osc_output.SendValue("/scratch/pos",	dPosition);	// Absolute position
		osc_output.SendValue("/scratch/TPS",	iTPS);		// Timecodes per second
		osc_output.SendValue("/scratch/speed",	dSpeed);
		osc_output.SendValue("/scratch/vol",	dVolumePeak);
		
/*
		strVolume	= "";
		for (int i =0; i<20;i++)
		{
			if (i <= VolumePeak)
				strVolume += ">";
			else
				strVolume += " ";
		}

		gotoxy(0,14);
		printf("%10.2fs %10.2f%%       %3i         %s\n", dPosition, 100.0 * scratch->GetSpeed(), iTPS, strVolume.c_str());
*/
		#ifdef WIN32
				Sleep(5);
		#else
				usleep(5000);	
		#endif

		if (bCloseApp)
			break;
	}

	delete scratch;
	return 0;
}